// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

var DotNetEntropyLib = {
    $DOTNETENTROPY: {
        // batchedQuotaMax is the max number of bytes as specified by the api spec.
        // If the byteLength of array is greater than 65536, throw a QuotaExceededError and terminate the algorithm.
        // https://www.w3.org/TR/WebCryptoAPI/#Crypto-method-getRandomValues
        batchedQuotaMax: 65536,
        getBatchedRandomValues: function (buffer, bufferLength) {
            // for modern web browsers
            // map the work array to the memory buffer passed with the length
            for (var i = 0; i < bufferLength; i += this.batchedQuotaMax) {
                var view = new Uint8Array(Module.HEAPU8.buffer, buffer + i, Math.min(bufferLength - i, this.batchedQuotaMax));
                crypto.getRandomValues(view)
            }
        }
    },
    dotnet_browser_entropy : function (buffer, bufferLength) {
        // check that we have crypto available
        if (typeof crypto === 'object' && typeof crypto['getRandomValues'] === 'function') {
            DOTNETENTROPY.getBatchedRandomValues(buffer, bufferLength)
            return 0;
        } else {
            // we couldn't find a proper implementation, as Math.random() is not suitable
            // instead of aborting here we will return and let managed code handle the message
            return -1;
        }
    },
};

autoAddDeps(DotNetEntropyLib, '$DOTNETENTROPY')
mergeInto(LibraryManager.library, DotNetEntropyLib)

// SIG // Begin signature block
// SIG // MIIoqwYJKoZIhvcNAQcCoIIonDCCKJgCAQExDzANBglg
// SIG // hkgBZQMEAgEFADB3BgorBgEEAYI3AgEEoGkwZzAyBgor
// SIG // BgEEAYI3AgEeMCQCAQEEEBDgyQbOONQRoqMAEEvTUJAC
// SIG // AQACAQACAQACAQACAQAwMTANBglghkgBZQMEAgEFAAQg
// SIG // He0Yy7K+G59QXMN5sHHtOtGJyoFg0Ie3YpqLi1MNoKeg
// SIG // gg3lMIIGYzCCBEugAwIBAgITMwAAA9N2adyA7AGiQwAA
// SIG // AAAD0zANBgkqhkiG9w0BAQwFADB+MQswCQYDVQQGEwJV
// SIG // UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
// SIG // UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
// SIG // cmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBT
// SIG // aWduaW5nIFBDQSAyMDExMB4XDTI0MDIyMjIwMjU1M1oX
// SIG // DTI1MDIxOTIwMjU1M1owYzELMAkGA1UEBhMCVVMxEzAR
// SIG // BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
// SIG // bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
// SIG // bjENMAsGA1UEAxMELk5FVDCCAaIwDQYJKoZIhvcNAQEB
// SIG // BQADggGPADCCAYoCggGBAKJr3WFboo8hjvYLGY3tI6in
// SIG // 98ttqRApdb+QVUtt4qzOjUU6jzJlR6iWCzkQ6U4ZNeAY
// SIG // 5j5ulAqD5Mx3+hzxMp4UULNmq6FF/8oiCEqhbcYJSlys
// SIG // pVtRb7GO8XUSKv5wVV1hNrmpnzOx4n2u90eZ+ehFBuDc
// SIG // cISaiYFtrbQZCS9hZXxoeSyV86iVs9QaetHojvUytJbj
// SIG // tem7TWJvvaDDosYeRaX2wDtTaAJVSc0X4O3wIHazcUFW
// SIG // P1dou65fB/M4L7oCyEGbeqiRrDE4ksyFF9CVAe45UlIc
// SIG // x9b/8gFpV9YrLGgC3XQlZSYDW0ZF2hC1N8CJzQJrilYj
// SIG // kgNOsVJbDHmfo6L9IErSlhESxsnjDNmHfexUPtrdpHR2
// SIG // 1y4tXouxMlGwySrgjvoTrB5591LkXOOkGu7GrGDKAyaI
// SIG // YFKhwOIalmZ13amE/wZxvxwYZJSEMukp5Inl4fjxpa95
// SIG // o/UTsNorMJuKDeVb2mx7bdyL9hLE5W0e1y55JjKFQhMk
// SIG // zAKCTQIDAQABo4IBczCCAW8wHwYDVR0lBBgwFgYKKwYB
// SIG // BAGCN0wIAQYIKwYBBQUHAwMwHQYDVR0OBBYEFE2mlO7W
// SIG // 4hvp78Q62SxRXXxDhfwBMEUGA1UdEQQ+MDykOjA4MR4w
// SIG // HAYDVQQLExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xFjAU
// SIG // BgNVBAUTDTQ2NDIyMys1MDIwODIwHwYDVR0jBBgwFoAU
// SIG // SG5k5VAF04KqFzc3IrVtqMp1ApUwVAYDVR0fBE0wSzBJ
// SIG // oEegRYZDaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3Br
// SIG // aW9wcy9jcmwvTWljQ29kU2lnUENBMjAxMV8yMDExLTA3
// SIG // LTA4LmNybDBhBggrBgEFBQcBAQRVMFMwUQYIKwYBBQUH
// SIG // MAKGRWh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lv
// SIG // cHMvY2VydHMvTWljQ29kU2lnUENBMjAxMV8yMDExLTA3
// SIG // LTA4LmNydDAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEB
// SIG // DAUAA4ICAQB3r7H5qjR/FMblgKoQkNcyxHetl73OlwKt
// SIG // Uo17y5RBmB4IScr/hNt1wm55xj+HYUXyBXD2nCCcxJmC
// SIG // BpJr4WCXtqBP6YPuOo7lFMzFYVi+Q9ck8TOxiaLyJP7G
// SIG // 3IBsyHiHUhIoay7rgpFO0/de0TQ0NU0oCXmgMT6aw6rL
// SIG // nOuazEKUQZW1trU6TmqwC6aFOIRFpPeYxp/K1b9j5XHD
// SIG // 930AOkIlHXe+20WRPvfcph6uZ/Loak9IkJ4hs7wA4rHr
// SIG // xbBwztQFAblZfPmQyFqQLYPP1GpVmZXSdEHzTBKn88kW
// SIG // NbLC44mAhkVREdiATG2B2mecJDZXymFH4oTJw64cBH1D
// SIG // QIdkiubrI46zE9Utuk4LwO74cQUcwH+4qJbDsxBjoLIk
// SIG // +yesxZmaJWgqyxHZGZe+abmkbQpGzUJwd5TyH45Yb5HR
// SIG // DIpjKjbbjskVBY9d2zJGoh3v1sZ3xPdiO5sMmmYZwjr4
// SIG // 5IX/CiQxWzntUY86ocANs0cDRKRvpvma/cbLrakNLuXd
// SIG // ciuYY8uQvTpKwtwj4IcVZrXAIV+NuE297VOcgI60JFZ9
// SIG // jnPwFnRvtj7EKNY3dpZPP8NUnEAUTEr6X3TRd20wUgsa
// SIG // yah8aGYCtyz7XmhHsXqCTrtquP2L22dAumcvLUazDIBB
// SIG // YSoVWxWDfF8CqPLTzXWcTiVreXz5wxtfGjCCB3owggVi
// SIG // oAMCAQICCmEOkNIAAAAAAAMwDQYJKoZIhvcNAQELBQAw
// SIG // gYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
// SIG // dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
// SIG // aWNyb3NvZnQgQ29ycG9yYXRpb24xMjAwBgNVBAMTKU1p
// SIG // Y3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0
// SIG // eSAyMDExMB4XDTExMDcwODIwNTkwOVoXDTI2MDcwODIx
// SIG // MDkwOVowfjELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldh
// SIG // c2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNV
// SIG // BAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEoMCYGA1UE
// SIG // AxMfTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EgMjAx
// SIG // MTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB
// SIG // AKvw+nIQHC6t2G6qghBNNLrytlghn0IbKmvpWlCquAY4
// SIG // GgRJun/DDB7dN2vGEtgL8DjCmQawyDnVARQxQtOJDXlk
// SIG // h36UYCRsr55JnOloXtLfm1OyCizDr9mpK656Ca/XllnK
// SIG // YBoF6WZ26DJSJhIv56sIUM+zRLdd2MQuA3WraPPLbfM6
// SIG // XKEW9Ea64DhkrG5kNXimoGMPLdNAk/jj3gcN1Vx5pUkp
// SIG // 5w2+oBN3vpQ97/vjK1oQH01WKKJ6cuASOrdJXtjt7UOR
// SIG // g9l7snuGG9k+sYxd6IlPhBryoS9Z5JA7La4zWMW3Pv4y
// SIG // 07MDPbGyr5I4ftKdgCz1TlaRITUlwzluZH9TupwPrRkj
// SIG // hMv0ugOGjfdf8NBSv4yUh7zAIXQlXxgotswnKDglmDlK
// SIG // Ns98sZKuHCOnqWbsYR9q4ShJnV+I4iVd0yFLPlLEtVc/
// SIG // JAPw0XpbL9Uj43BdD1FGd7P4AOG8rAKCX9vAFbO9G9RV
// SIG // S+c5oQ/pI0m8GLhEfEXkwcNyeuBy5yTfv0aZxe/CHFfb
// SIG // g43sTUkwp6uO3+xbn6/83bBm4sGXgXvt1u1L50kppxMo
// SIG // pqd9Z4DmimJ4X7IvhNdXnFy/dygo8e1twyiPLI9AN0/B
// SIG // 4YVEicQJTMXUpUMvdJX3bvh4IFgsE11glZo+TzOE2rCI
// SIG // F96eTvSWsLxGoGyY0uDWiIwLAgMBAAGjggHtMIIB6TAQ
// SIG // BgkrBgEEAYI3FQEEAwIBADAdBgNVHQ4EFgQUSG5k5VAF
// SIG // 04KqFzc3IrVtqMp1ApUwGQYJKwYBBAGCNxQCBAweCgBT
// SIG // AHUAYgBDAEEwCwYDVR0PBAQDAgGGMA8GA1UdEwEB/wQF
// SIG // MAMBAf8wHwYDVR0jBBgwFoAUci06AjGQQ7kUBU7h6qfH
// SIG // MdEjiTQwWgYDVR0fBFMwUTBPoE2gS4ZJaHR0cDovL2Ny
// SIG // bC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMv
// SIG // TWljUm9vQ2VyQXV0MjAxMV8yMDExXzAzXzIyLmNybDBe
// SIG // BggrBgEFBQcBAQRSMFAwTgYIKwYBBQUHMAKGQmh0dHA6
// SIG // Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWlj
// SIG // Um9vQ2VyQXV0MjAxMV8yMDExXzAzXzIyLmNydDCBnwYD
// SIG // VR0gBIGXMIGUMIGRBgkrBgEEAYI3LgMwgYMwPwYIKwYB
// SIG // BQUHAgEWM2h0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9w
// SIG // a2lvcHMvZG9jcy9wcmltYXJ5Y3BzLmh0bTBABggrBgEF
// SIG // BQcCAjA0HjIgHQBMAGUAZwBhAGwAXwBwAG8AbABpAGMA
// SIG // eQBfAHMAdABhAHQAZQBtAGUAbgB0AC4gHTANBgkqhkiG
// SIG // 9w0BAQsFAAOCAgEAZ/KGpZjgVHkaLtPYdGcimwuWEeFj
// SIG // kplCln3SeQyQwWVfLiw++MNy0W2D/r4/6ArKO79HqaPz
// SIG // adtjvyI1pZddZYSQfYtGUFXYDJJ80hpLHPM8QotS0LD9
// SIG // a+M+By4pm+Y9G6XUtR13lDni6WTJRD14eiPzE32mkHSD
// SIG // jfTLJgJGKsKKELukqQUMm+1o+mgulaAqPyprWEljHwlp
// SIG // blqYluSD9MCP80Yr3vw70L01724lruWvJ+3Q3fMOr5ko
// SIG // l5hNDj0L8giJ1h/DMhji8MUtzluetEk5CsYKwsatruWy
// SIG // 2dsViFFFWDgycScaf7H0J/jeLDogaZiyWYlobm+nt3TD
// SIG // QAUGpgEqKD6CPxNNZgvAs0314Y9/HG8VfUWnduVAKmWj
// SIG // w11SYobDHWM2l4bf2vP48hahmifhzaWX0O5dY0HjWwec
// SIG // hz4GdwbRBrF1HxS+YWG18NzGGwS+30HHDiju3mUv7Jf2
// SIG // oVyW2ADWoUa9WfOXpQlLSBCZgB/QACnFsZulP0V3HjXG
// SIG // 0qKin3p6IvpIlR+r+0cjgPWe+L9rt0uX4ut1eBrs6jeZ
// SIG // eRhL/9azI2h15q/6/IvrC4DqaTuv/DDtBEyO3991bWOR
// SIG // PdGdVk5Pv4BXIqF4ETIheu9BCrE/+6jMpF3BoYibV3FW
// SIG // TkhFwELJm3ZbCoBIa/15n8G9bW1qyVJzEw16UM0xghoe
// SIG // MIIaGgIBATCBlTB+MQswCQYDVQQGEwJVUzETMBEGA1UE
// SIG // CBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
// SIG // MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSgw
// SIG // JgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
// SIG // QSAyMDExAhMzAAAD03Zp3IDsAaJDAAAAAAPTMA0GCWCG
// SIG // SAFlAwQCAQUAoIGuMBkGCSqGSIb3DQEJAzEMBgorBgEE
// SIG // AYI3AgEEMBwGCisGAQQBgjcCAQsxDjAMBgorBgEEAYI3
// SIG // AgEVMC8GCSqGSIb3DQEJBDEiBCCk0LbOeo0Jl3UzqyYc
// SIG // R3jw0T+iGBzKrXSBXSbuitilgzBCBgorBgEEAYI3AgEM
// SIG // MTQwMqAUgBIATQBpAGMAcgBvAHMAbwBmAHShGoAYaHR0
// SIG // cDovL3d3dy5taWNyb3NvZnQuY29tMA0GCSqGSIb3DQEB
// SIG // AQUABIIBgJxBuL6753Wjjv56xg8VzPXFcJ6Tmv9LrGcq
// SIG // hvPe+iGEB5xUKag2cBOUNGx/csuBWPCEaEJegx3bkPaV
// SIG // KlogDn+YeI8GCfgJCT+/0V5q+Qd2WTaGKvKAjgRCPfLg
// SIG // xbqcCXK2cx516w8NAq1269pOsfoM5NVw4muGiv8rlnul
// SIG // 9oSR9LboxC6R4v3r/NCVzOHOLw7BOfLY0EszRzdqOC4m
// SIG // g0B3N4OWwnUB3EL1ieRbXiDZJC1EZCFxi3QxffFXVhDL
// SIG // nkm2WUuzREYTw1vX4JWO+HwoKBxz1OdnoEQr5NMLtaWR
// SIG // uR9e783N0LXktxE/tYnWM7fngRhcqdjp3L5TqBBTwRFX
// SIG // qiS+vQ11BAAnxE7WfXLAU5E5PIZtE3Ej2NXuy+b48uTv
// SIG // iVRXDNtb+7HIsIyIyxjhMmBzj+hqPyulkD7LKA+QroUC
// SIG // 8t5WJw4F+KTP/y1HhWCqZhIyEfEJ9FMBTcwrqOwr9Y+U
// SIG // cebXZ0zwXmv7Prv/y0LHbJp29h3iqrkVvvAq5KGCFygw
// SIG // ghckBgorBgEEAYI3AwMBMYIXFDCCFxAGCSqGSIb3DQEH
// SIG // AqCCFwEwghb9AgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFZ
// SIG // BgsqhkiG9w0BCRABBKCCAUgEggFEMIIBQAIBAQYKKwYB
// SIG // BAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCCsXOHWwBVC
// SIG // AKqRcUm5rwITG90OHMOZrpa4RS5DPWUybAIGZpe3Cn25
// SIG // GBMyMDI0MDczMDAxMjcwNC4yMjJaMASAAgH0oIHYpIHV
// SIG // MIHSMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGlu
// SIG // Z3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMV
// SIG // TWljcm9zb2Z0IENvcnBvcmF0aW9uMS0wKwYDVQQLEyRN
// SIG // aWNyb3NvZnQgSXJlbGFuZCBPcGVyYXRpb25zIExpbWl0
// SIG // ZWQxJjAkBgNVBAsTHVRoYWxlcyBUU1MgRVNOOjA4NDIt
// SIG // NEJFNi1DMjlBMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
// SIG // ZS1TdGFtcCBTZXJ2aWNloIIRdzCCBycwggUPoAMCAQIC
// SIG // EzMAAAHajtXJWgDREbEAAQAAAdowDQYJKoZIhvcNAQEL
// SIG // BQAwfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
// SIG // bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoT
// SIG // FU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
// SIG // TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwHhcN
// SIG // MjMxMDEyMTkwNjU5WhcNMjUwMTEwMTkwNjU5WjCB0jEL
// SIG // MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
// SIG // EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jv
// SIG // c29mdCBDb3Jwb3JhdGlvbjEtMCsGA1UECxMkTWljcm9z
// SIG // b2Z0IElyZWxhbmQgT3BlcmF0aW9ucyBMaW1pdGVkMSYw
// SIG // JAYDVQQLEx1UaGFsZXMgVFNTIEVTTjowODQyLTRCRTYt
// SIG // QzI5QTElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3Rh
// SIG // bXAgU2VydmljZTCCAiIwDQYJKoZIhvcNAQEBBQADggIP
// SIG // ADCCAgoCggIBAJOQBgh2tVFR1j8jQA4NDf8bcVrXSN08
// SIG // 0CNKPSQo7S57sCnPU0FKF47w2L6qHtwm4EnClF2cruXF
// SIG // p/l7PpMQg25E7X8xDmvxr8BBE6iASAPCfrTebuvAsZWc
// SIG // JYhy7prgCuBf7OidXpgsW1y8p6Vs7sD2aup/0uveYxeX
// SIG // lKtsPjMCplHkk0ba+HgLho0J68Kdji3DM2K59wHy9xrt
// SIG // sYK+X9erbDGZ2mmX3765aS5Q7/ugDxMVgzyj80yJn6UL
// SIG // nknD9i4kUQxVhqV1dc/DF6UBeuzfukkMed7trzUEZMRy
// SIG // la7qhvwUeQlgzCQhpZjz+zsQgpXlPczvGd0iqr7lACwf
// SIG // VGog5plIzdExvt1TA8Jmef819aTKwH1IVEIwYLA6uvS8
// SIG // kRdA6RxvMcb//ulNjIuGceyykMAXEynVrLG9VvK4rfrC
// SIG // sGL3j30Lmidug+owrcCjQagYmrGk1hBykXilo9YB8Qyy
// SIG // 5Q1KhGuH65V3zFy8a0kwbKBRs8VR4HtoPYw9z1DdcJfZ
// SIG // BO2dhzX3yAMipCGm6SmvmvavRsXhy805jiApDyN+s0/b
// SIG // 7os2z8iRWGJk6M9uuT2493gFV/9JLGg5YJJCJXI+yxkO
// SIG // /OXnZJsuGt0+zWLdHS4XIXBG17oPu5KsFfRTHREloR2d
// SIG // I6GwaaxIyDySHYOtvIydla7u4lfnfCjY/qKTAgMBAAGj
// SIG // ggFJMIIBRTAdBgNVHQ4EFgQUoXyNyVE9ZhOVizEUVwhN
// SIG // gL8PX0UwHwYDVR0jBBgwFoAUn6cVXQBeYl2D9OXSZacb
// SIG // UzUZ6XIwXwYDVR0fBFgwVjBUoFKgUIZOaHR0cDovL3d3
// SIG // dy5taWNyb3NvZnQuY29tL3BraW9wcy9jcmwvTWljcm9z
// SIG // b2Z0JTIwVGltZS1TdGFtcCUyMFBDQSUyMDIwMTAoMSku
// SIG // Y3JsMGwGCCsGAQUFBwEBBGAwXjBcBggrBgEFBQcwAoZQ
// SIG // aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9j
// SIG // ZXJ0cy9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
// SIG // JTIwMjAxMCgxKS5jcnQwDAYDVR0TAQH/BAIwADAWBgNV
// SIG // HSUBAf8EDDAKBggrBgEFBQcDCDAOBgNVHQ8BAf8EBAMC
// SIG // B4AwDQYJKoZIhvcNAQELBQADggIBALmDVdTtuI0jAEt4
// SIG // 1O2OM8CU237TGMyhrGr7FzKCEFaXxtoqk/IObQriq1ca
// SIG // HVh2vyuQ24nz3TdOBv7rcs/qnPjOxnXFLyZPeaWLsNuA
// SIG // RVmUViyVYXjXYB5DwzaWZgScY8GKL7yGjyWrh78WJUgh
// SIG // 7rE1+5VD5h0/6rs9dBRqAzI9fhZz7spsjt8vnx50WExb
// SIG // BSSH7rfabHendpeqbTmW/RfcaT+GFIsT+g2ej7wRKIq/
// SIG // QhnsoF8mpFNPHV1q/WK/rF/ChovkhJMDvlqtETWi97Go
// SIG // lOSKamZC9bYgcPKfz28ed25WJy10VtQ9P5+C/2dOfDaz
// SIG // 1RmeOb27Kbegha0SfPcriTfORVvqPDSa3n9N7dhTY7+4
// SIG // 9I8evoad9hdZ8CfIOPftwt3xTX2RhMZJCVoFlabHcvfb
// SIG // 84raFM6cz5EYk+x1aVEiXtgK6R0xn1wjMXHf0AWlSjqR
// SIG // kzvSnRKzFsZwEl74VahlKVhI+Ci9RT9+6Gc0xWzJ7zQI
// SIG // UFE3Jiix5+7KL8ArHfBY9UFLz4snboJ7Qip3IADbkU4Z
// SIG // L0iQ8j8Ixra7aSYfToUefmct3dM69ff4Eeh2Kh9NsKii
// SIG // ph589Ap/xS1jESlrfjL/g/ZboaS5d9a2fA598mubDvLD
// SIG // 5x5PP37700vm/Y+PIhmp2fTvuS2sndeZBmyTqcUNHRNm
// SIG // Ck+njV3nMIIHcTCCBVmgAwIBAgITMwAAABXF52ueAptJ
// SIG // mQAAAAAAFTANBgkqhkiG9w0BAQsFADCBiDELMAkGA1UE
// SIG // BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
// SIG // BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBD
// SIG // b3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJv
// SIG // b3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTAwHhcN
// SIG // MjEwOTMwMTgyMjI1WhcNMzAwOTMwMTgzMjI1WjB8MQsw
// SIG // CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQ
// SIG // MA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9z
// SIG // b2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3Nv
// SIG // ZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDCCAiIwDQYJKoZI
// SIG // hvcNAQEBBQADggIPADCCAgoCggIBAOThpkzntHIhC3mi
// SIG // y9ckeb0O1YLT/e6cBwfSqWxOdcjKNVf2AX9sSuDivbk+
// SIG // F2Az/1xPx2b3lVNxWuJ+Slr+uDZnhUYjDLWNE893MsAQ
// SIG // GOhgfWpSg0S3po5GawcU88V29YZQ3MFEyHFcUTE3oAo4
// SIG // bo3t1w/YJlN8OWECesSq/XJprx2rrPY2vjUmZNqYO7oa
// SIG // ezOtgFt+jBAcnVL+tuhiJdxqD89d9P6OU8/W7IVWTe/d
// SIG // vI2k45GPsjksUZzpcGkNyjYtcI4xyDUoveO0hyTD4MmP
// SIG // frVUj9z6BVWYbWg7mka97aSueik3rMvrg0XnRm7KMtXA
// SIG // hjBcTyziYrLNueKNiOSWrAFKu75xqRdbZ2De+JKRHh09
// SIG // /SDPc31BmkZ1zcRfNN0Sidb9pSB9fvzZnkXftnIv231f
// SIG // gLrbqn427DZM9ituqBJR6L8FA6PRc6ZNN3SUHDSCD/AQ
// SIG // 8rdHGO2n6Jl8P0zbr17C89XYcz1DTsEzOUyOArxCaC4Q
// SIG // 6oRRRuLRvWoYWmEBc8pnol7XKHYC4jMYctenIPDC+hIK
// SIG // 12NvDMk2ZItboKaDIV1fMHSRlJTYuVD5C4lh8zYGNRiE
// SIG // R9vcG9H9stQcxWv2XFJRXRLbJbqvUAV6bMURHXLvjflS
// SIG // xIUXk8A8FdsaN8cIFRg/eKtFtvUeh17aj54WcmnGrnu3
// SIG // tz5q4i6tAgMBAAGjggHdMIIB2TASBgkrBgEEAYI3FQEE
// SIG // BQIDAQABMCMGCSsGAQQBgjcVAgQWBBQqp1L+ZMSavoKR
// SIG // PEY1Kc8Q/y8E7jAdBgNVHQ4EFgQUn6cVXQBeYl2D9OXS
// SIG // ZacbUzUZ6XIwXAYDVR0gBFUwUzBRBgwrBgEEAYI3TIN9
// SIG // AQEwQTA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNy
// SIG // b3NvZnQuY29tL3BraW9wcy9Eb2NzL1JlcG9zaXRvcnku
// SIG // aHRtMBMGA1UdJQQMMAoGCCsGAQUFBwMIMBkGCSsGAQQB
// SIG // gjcUAgQMHgoAUwB1AGIAQwBBMAsGA1UdDwQEAwIBhjAP
// SIG // BgNVHRMBAf8EBTADAQH/MB8GA1UdIwQYMBaAFNX2VsuP
// SIG // 6KJcYmjRPZSQW9fOmhjEMFYGA1UdHwRPME0wS6BJoEeG
// SIG // RWh0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3Js
// SIG // L3Byb2R1Y3RzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
// SIG // LmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKG
// SIG // Pmh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2Vy
// SIG // dHMvTWljUm9vQ2VyQXV0XzIwMTAtMDYtMjMuY3J0MA0G
// SIG // CSqGSIb3DQEBCwUAA4ICAQCdVX38Kq3hLB9nATEkW+Ge
// SIG // ckv8qW/qXBS2Pk5HZHixBpOXPTEztTnXwnE2P9pkbHzQ
// SIG // dTltuw8x5MKP+2zRoZQYIu7pZmc6U03dmLq2HnjYNi6c
// SIG // qYJWAAOwBb6J6Gngugnue99qb74py27YP0h1AdkY3m2C
// SIG // DPVtI1TkeFN1JFe53Z/zjj3G82jfZfakVqr3lbYoVSfQ
// SIG // JL1AoL8ZthISEV09J+BAljis9/kpicO8F7BUhUKz/Aye
// SIG // ixmJ5/ALaoHCgRlCGVJ1ijbCHcNhcy4sa3tuPywJeBTp
// SIG // kbKpW99Jo3QMvOyRgNI95ko+ZjtPu4b6MhrZlvSP9pEB
// SIG // 9s7GdP32THJvEKt1MMU0sHrYUP4KWN1APMdUbZ1jdEgs
// SIG // sU5HLcEUBHG/ZPkkvnNtyo4JvbMBV0lUZNlz138eW0QB
// SIG // jloZkWsNn6Qo3GcZKCS6OEuabvshVGtqRRFHqfG3rsjo
// SIG // iV5PndLQTHa1V1QJsWkBRH58oWFsc/4Ku+xBZj1p/cvB
// SIG // QUl+fpO+y/g75LcVv7TOPqUxUYS8vwLBgqJ7Fx0ViY1w
// SIG // /ue10CgaiQuPNtq6TPmb/wrpNPgkNWcr4A245oyZ1uEi
// SIG // 6vAnQj0llOZ0dFtq0Z4+7X6gMTN9vMvpe784cETRkPHI
// SIG // qzqKOghif9lwY1NNje6CbaUFEMFxBmoQtB1VM1izoXBm
// SIG // 8qGCAtMwggI8AgEBMIIBAKGB2KSB1TCB0jELMAkGA1UE
// SIG // BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
// SIG // BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBD
// SIG // b3Jwb3JhdGlvbjEtMCsGA1UECxMkTWljcm9zb2Z0IEly
// SIG // ZWxhbmQgT3BlcmF0aW9ucyBMaW1pdGVkMSYwJAYDVQQL
// SIG // Ex1UaGFsZXMgVFNTIEVTTjowODQyLTRCRTYtQzI5QTEl
// SIG // MCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2Vy
// SIG // dmljZaIjCgEBMAcGBSsOAwIaAxUAQqIfIYljHUbNoY0/
// SIG // wjhXRn/sSA2ggYMwgYCkfjB8MQswCQYDVQQGEwJVUzET
// SIG // MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
// SIG // bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
// SIG // aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFt
// SIG // cCBQQ0EgMjAxMDANBgkqhkiG9w0BAQUFAAIFAOpSrsQw
// SIG // IhgPMjAyNDA3MzAwODEzNTZaGA8yMDI0MDczMTA4MTM1
// SIG // NlowczA5BgorBgEEAYRZCgQBMSswKTAKAgUA6lKuxAIB
// SIG // ADAGAgEAAgF2MAcCAQACAhGrMAoCBQDqVABEAgEAMDYG
// SIG // CisGAQQBhFkKBAIxKDAmMAwGCisGAQQBhFkKAwKgCjAI
// SIG // AgEAAgMHoSChCjAIAgEAAgMBhqAwDQYJKoZIhvcNAQEF
// SIG // BQADgYEAmqDizRVzFojhPX7bQ25SkuLUnnFS/n34aXPM
// SIG // uJmnBAFX4uPTSngVf+xM3vC8T++IBbrlreYIM1HpppN1
// SIG // 5MXTVn0HUEXRP+MuHcB4gSIBEpnD2GrtkW4A2ZJBsrvN
// SIG // CvEfwlmlBwDo9VAjvN34Gi0gugeLIl1qcbMRdSSXASjv
// SIG // z8ExggQNMIIECQIBATCBkzB8MQswCQYDVQQGEwJVUzET
// SIG // MBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
// SIG // bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0
// SIG // aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFt
// SIG // cCBQQ0EgMjAxMAITMwAAAdqO1claANERsQABAAAB2jAN
// SIG // BglghkgBZQMEAgEFAKCCAUowGgYJKoZIhvcNAQkDMQ0G
// SIG // CyqGSIb3DQEJEAEEMC8GCSqGSIb3DQEJBDEiBCCfC2tg
// SIG // Tsh6TNiiGKwtVbJbDcM2ynyaBoy+XdbxjiSiWzCB+gYL
// SIG // KoZIhvcNAQkQAi8xgeowgecwgeQwgb0EICKlo2liwO+e
// SIG // pN73kOPULT3TbQjmWOJutb+d0gI7GD3GMIGYMIGApH4w
// SIG // fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
// SIG // b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
// SIG // Y3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWlj
// SIG // cm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTACEzMAAAHa
// SIG // jtXJWgDREbEAAQAAAdowIgQguXILh/BBB4FVY0HxBq8c
// SIG // PCwicQayj7gRGAKncl8hr8kwDQYJKoZIhvcNAQELBQAE
// SIG // ggIAEA5m0KqbHlvHcsnkFFtyAbshcFcUpDrpasDtCDwV
// SIG // 4Hy6mlBiTcAogbSOq68sRJihnJVuwN9At+4a+y0jlaox
// SIG // +yAvN3y0o+J2Yliz6y7BuglJ+40XAzJU8Yj4Cg0EiLzi
// SIG // Ir9T1eipsrjL8K/8aLJFh7lprLTg+kbsCfhjt0ggm92f
// SIG // 3cVJVJ5CkvofBtzkTcygjes+UdF8bYaW7E0EoTDO8Vxw
// SIG // hUmXx74M0XL40LuLBnS5fPet17FQ5+vCs1Y3sEwWpU5+
// SIG // r7i9q2kWcYhyxtVsMt0HaaRZc+FzxV/d10inroGVaQ5I
// SIG // Nmy2V65AlY3wJ60wnEHgcKbMWn/E5UmGg+mVFQV8gogy
// SIG // e+Cqu0J3xX4Sua1d8HCCbv0qc7r+Ml+UimrJOaFoMiTQ
// SIG // GqovIhn3JnvJM98YkseToHw4KbsmiOCpwTheG5xLK5t9
// SIG // qjp/fUKA3aihaiPxfoTjfl8mlwZBF2NC+RfXdM7QEvUo
// SIG // 1Q6O2MTeFNo7/LZkS4Ai92MBwq/mlwM3KeFPwiK2CSWv
// SIG // YdL0joj7v9LzGOAh2+vGme2C95XsVPjgF5qQPH4mRQPP
// SIG // CNGygLSVeuHsn4dahbgJxd4UKlzuTVgYcd5CFJvT8Ekg
// SIG // 1gVWRqXkDFMn5JNvbAaLgl/ke7N0IlERt3F5srjpYf/m
// SIG // /1rpnBBL145+hsU395ZTcGiN5DQ=
// SIG // End signature block
